from enum import Enum

from flask_admin.model import filters
from bukuserver import _l, _key


class BookmarkField(Enum):
    ID = 0
    URL = 1
    TITLE = 2
    TAGS = 3
    DESCRIPTION = 4


def equal_func(query, value, index):
    return filter(lambda x: x[index] == value, query)


def not_equal_func(query, value, index):
    return filter(lambda x: x[index] != value, query)


def contains_func(query, value, index):
    return filter(lambda x: value in x[index], query)


def not_contains_func(query, value, index):
    return filter(lambda x: value not in x[index], query)


def greater_func(query, value, index):
    return filter(lambda x: x[index] > value, query)


def smaller_func(query, value, index):
    return filter(lambda x: x[index] < value, query)


def in_list_func(query, value, index):
    return filter(lambda x: x[index] in value, query)


def not_in_list_func(query, value, index):
    return filter(lambda x: x[index] not in value, query)


def top_x_func(query, value, index):
    items = sorted(set(x[index] for x in query), reverse=True)
    top_x = items[:value]
    return filter(lambda x: x[index] in top_x, query)


def bottom_x_func(query, value, index):
    items = sorted(set(x[index] for x in query), reverse=False)
    top_x = items[:value]
    return filter(lambda x: x[index] in top_x, query)


class FilterType(Enum):

    EQUAL = {'func': equal_func, 'text': _l('equals')}
    NOT_EQUAL = {'func': not_equal_func, 'text': _l('not equals')}
    CONTAINS = {'func': contains_func, 'text': _l('contains')}
    NOT_CONTAINS = {'func': not_contains_func, 'text': _l('not contains')}
    GREATER = {'func': greater_func, 'text': _l('greater than')}
    SMALLER = {'func': smaller_func, 'text': _l('smaller than')}
    IN_LIST = {'func': in_list_func, 'text': _l('in list')}
    NOT_IN_LIST = {'func': not_in_list_func, 'text': _l('not in list')}
    TOP_X = {'func': top_x_func, 'text': _l('top X')}
    BOTTOM_X = {'func': bottom_x_func, 'text': _l('bottom X')}


class BaseFilter(filters.BaseFilter):

    def operation(self):
        return getattr(self, 'operation_text')

    def apply(self, query, value):
        return getattr(self, 'apply_func')(query, value, getattr(self, 'index'))


class TagBaseFilter(BaseFilter):

    def __init__(
            self,
            name,
            operation_text=None,
            apply_func=None,
            filter_type=None,
            options=None,
            data_type=None):
        try:
            self.index = ['name', 'usage_count'].index(name)
        except ValueError as e:
            raise ValueError(f'name: {name}') from e
        self.filter_type = filter_type
        if filter_type:
            self.apply_func = filter_type.value['func']
            self.operation_text = filter_type.value['text']
        else:
            self.apply_func = apply_func
            self.operation_text = operation_text
        if _key(self.operation_text) in ('in list', 'not in list'):
            super().__init__(name, options, data_type='select2-tags')
        else:
            super().__init__(name, options, data_type)

    def clean(self, value):
        on_list = self.filter_type in (FilterType.IN_LIST, FilterType.NOT_IN_LIST)
        if on_list and self.name == 'usage_count':
            value = [int(v.strip()) for v in value.split(',') if v.strip()]
        elif on_list:
            value = [v.strip() for v in value.split(',') if v.strip()]
        elif self.name == 'usage_count':
            value = int(value)
            if self.filter_type in (FilterType.TOP_X, FilterType.BOTTOM_X) and value < 1:
                raise ValueError
        if isinstance(value, str):
            return value.strip()
        return value


class BookmarkOrderFilter(BaseFilter):
    DIR_LIST = [('asc', _l('natural')), ('desc', _l('reversed'))]
    FIELDS = ['index', 'url', 'netloc', 'title', 'description', 'tags']

    def __init__(self, field, *args, **kwargs):
        self.field = field
        super().__init__('order', *args, options=self.DIR_LIST, **kwargs)

    def operation(self):
        return _l(f'by {self.field}')

    def apply(self, query, value):
        return query

    @staticmethod
    def value(filters, values):
        return [('-' if value == 'desc' else '+') + filters[idx].field
                for idx, key, value in values if key == 'order']


class BookmarkBukuFilter(BaseFilter):
    KEYS = {
        'markers': 'markers',
        'all_keywords': 'match all',
        'deep': 'deep',
        'regex': 'regex',
    }

    def __init__(self, *args, **kwargs):
        self.params = {key: kwargs.pop(key, False) for key in self.KEYS}
        super().__init__('buku', *args, **kwargs)

    def operation(self):
        parts = ', '.join(v for k, v in self.KEYS.items() if self.params[k])
        return _l(f'search{parts and " " + parts}')

    def apply(self, query, value):
        return query


class BookmarkBaseFilter(BaseFilter):

    def __init__(
            self,
            name,
            operation_text=None,
            apply_func=None,
            filter_type=None,
            options=None,
            data_type=None):
        bm_fields_dict = {x.name.lower(): x.value for x in BookmarkField}
        if name in bm_fields_dict:
            self.index = bm_fields_dict[name]
        else:
            raise ValueError(f'name: {name}')
        self.filter_type = None
        if filter_type:
            self.apply_func = filter_type.value['func']
            self.operation_text = filter_type.value['text']
        else:
            self.apply_func = apply_func
            self.operation_text = operation_text
        if _key(self.operation_text) in ('in list', 'not in list'):
            super().__init__(name, options, data_type='select2-tags')
        else:
            super().__init__(name, options, data_type)

    def clean(self, value):
        on_list = _key(self.operation_text) in ('in list', 'not in list')
        if on_list and self.name == BookmarkField.ID.name.lower():
            value = [int(v.strip()) for v in value.split(',') if v.strip()]
        elif on_list:
            value = [v.strip() for v in value.split(',') if v.strip()]
        elif self.name == BookmarkField.ID.name.lower():
            value = int(value)
            if self.filter_type in (FilterType.TOP_X, FilterType.BOTTOM_X) and value < 1:
                raise ValueError
        if isinstance(value, str):
            return value.strip()
        return value


class BookmarkTagNumberEqualFilter(BookmarkBaseFilter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def apply_func(query, value, index):
            for item in query:
                tags = [tag for tag in item[index].split(',') if tag]
                if len(tags) == value:
                    yield item

        self.apply_func = apply_func

    def clean(self, value):
        value = int(value)
        if value < 0:
            raise ValueError
        return value


class BookmarkTagNumberGreaterFilter(BookmarkTagNumberEqualFilter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def apply_func(query, value, index):
            for item in query:
                tags = [tag for tag in item[index].split(',') if tag]
                if len(tags) > value:
                    yield item

        self.apply_func = apply_func


class BookmarkTagNumberNotEqualFilter(BookmarkTagNumberEqualFilter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def apply_func(query, value, index):
            for item in query:
                tags = [tag for tag in item[index].split(',') if tag]
                if len(tags) != value:
                    yield item

        self.apply_func = apply_func


class BookmarkTagNumberSmallerFilter(BookmarkBaseFilter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def apply_func(query, value, index):
            for item in query:
                tags = [tag for tag in item[index].split(',') if tag]
                if len(tags) < value:
                    yield item

        self.apply_func = apply_func

    def clean(self, value):
        value = int(value)
        if value < 1:
            raise ValueError
        return value



import os
import re
from babel.messages.frontend import CommandLineInterface as pybabel
try:
    from buku import __version__
except ImportError:
    __version__ = None

DOMAIN = 'messages'
DIR = os.path.dirname(__file__)
BUKUSERVER = os.path.dirname(DIR)
MAPPING = os.path.join(DIR, 'babel.cfg')
TEMPLATE = os.path.join(DIR, 'messages.pot')
CUSTOM = os.path.join(DIR, 'messages_custom.pot')

_EOL, _STR_EOL, _G_STR_EOL = r'$\r?\n?', r'"[^\r\n]*"$\r?\n?', r'"([^\r\n]*)"$\r?\n?'
STRINGS = {
    r'\bPROJECT VERSION\b': __version__ or '???',
    r'(?<=for )PROJECT\b': 'bukuserver',
    r'(?<=as the )PROJECT(?= project)': 'buku',
    r'\bORGANIZATION\b': 'buku',
    f'^# FIRST AUTHOR <EMAIL@ADDRESS>, [0-9]+.{_EOL}': '',
    f'^#, fuzzy{_EOL}': '',
    r'(?<=^"POT-Creation-Date: ).*(?=\\n"$)': '2024-09-12 00:00+0000',  # avoid git updates of unchanged translations
}
OLD_BLANK = re.compile(f'^(?:#~ msgctxt {_STR_EOL})?#~ msgid {_STR_EOL}#~ msgstr ""{_EOL}(?:{_EOL})?', re.MULTILINE)
OLD = re.compile(f'^(?:#~ msgctxt {_G_STR_EOL})?#~ msgid {_G_STR_EOL}#~ msgstr {_G_STR_EOL}(?:{_EOL})?', re.MULTILINE)

def replace_obsolete(text):
    '''Removes *blank* obsolete entries, and restores re-added *blank* values from old obsolete ones'''
    text = re.sub(OLD_BLANK, '', text)
    for obsolete in re.finditer(OLD, text):
        _ctxt, _id, _str = obsolete.groups()
        ctxt_re = ('' if _ctxt is None else f'msgctxt "{re.escape(_ctxt)}"{_EOL}')
        if m := re.search(f'^{ctxt_re}msgid "{re.escape(_id)}"{_EOL}msgstr "()"{_EOL}', text, re.MULTILINE):
            text = (text[:m.start(1)] + _str + text[m.end(1):]).replace(obsolete.group(0), '', 1)
    return text

def translations_generate():
    '''Generates and patches the messages.pot template file'''
    pybabel().run(['', 'extract', '--no-wrap', f'--mapping-file={MAPPING}',
                   '--keywords=_ _l _p:1c,2 _lp:1c,2 lazy_gettext', f'--output-file={TEMPLATE}', BUKUSERVER])
    print(f'patching PO template file at {TEMPLATE}')
    with open(TEMPLATE, encoding='utf-8') as fin:
        text = fin.read()
    for k, v in STRINGS.items():
        text = re.sub(k, v, text, count=1, flags=re.MULTILINE)
    with open(CUSTOM, encoding='utf-8') as fin:
        with open(TEMPLATE, 'w', encoding='utf-8') as fout:
            fout.write(text + fin.read())

def translations_update(new_locales=[], generate=True, domain=DOMAIN, fuzzy=False):
    '''Updates all existing translations (*/LC_MESSAGES/messages.po) based on messages.pot'''
    generate and translations_generate()
    command = (['', 'update', '--no-wrap'] + ([] if fuzzy else ['--no-fuzzy-matching']) +
               [f'--domain={domain}', f'--input-file={TEMPLATE}', f'--output-dir={DIR}'])
    pybabel().run(command)
    for locale in new_locales:
        try:  # trying an update first, to prevent clearing an existing translation
            pybabel().run(command + ['--init-missing', f'--locale={locale}'])
        except FileNotFoundError:
            pybabel().run(['', 'init', '--no-wrap', f'--domain={domain}', f'--input-file={TEMPLATE}',
                           f'--output-dir={DIR}', f'--locale={locale}'])
    # handling obsolete entries (removing blank and restoring re-added keys)
    for locale in os.listdir(DIR):
        filename = os.path.join(DIR, locale, 'LC_MESSAGES', f'{domain}.po')
        if os.path.isfile(filename) and os.access(filename, os.W_OK):
            with open(filename, encoding='utf-8') as fin:
                text = fin.read()
            if text != (stripped := replace_obsolete(text)):
                print(f'processing obsolete entries from catalog {filename}')
                with open(filename, 'w', encoding='utf-8') as fout:
                    fout.write(stripped)

def translations_compile(update=False, generate=True, domain=DOMAIN, new_locales=[], fuzzy=False):
    '''Compiles all existing translations'''
    update and translations_update(generate=generate, domain=DOMAIN, new_locales=new_locales, fuzzy=fuzzy)
    pybabel().run(['', 'compile', f'--domain={domain}', f'--directory={DIR}'])



import json
from buku import import_firefox_json


def test_load_from_empty():
    """test method."""
    # Arrange
    data = json.loads("{}")

    # Act
    items = import_firefox_json(data)

    # Assert
    count = sum(1 for _ in items)
    assert count == 0

def test_load_full_entry():
    """test method."""

    # Arrange
    data = json.loads("""
        {
            "title" : "main",
            "typeCode": 2,
            "children": [
                {
                "title" : "title",
                "typeCode": 2,
                "children": [
                    {
                      "dateAdded": 1269200039653000,
                      "guid": "xxxydfalkj",
                      "id": 113,
                      "index": 0,
                      "lastModified": 1305978154986000,
                      "title": "entry title",
                      "type": "text/x-moz-place",
                      "typeCode": 1,
                      "tags" : "x,y",
                      "uri": "http://uri.com/abc?234&536",
                      "annos" : [{
                             "name": "bookmarkProperties/description",
                             "value": "desc"
                       }]
                    }]
             }]
        }""")

    # Act
    items = import_firefox_json(data)

    # Assert
    result = []
    for item in items:
        result.append(item)

    assert len(result) == 1
    assert result[0][0] == 'http://uri.com/abc?234&536'
    assert result[0][1] == 'entry title'
    assert result[0][2] == ',x,y,'
    assert result[0][3] == 'desc'


def test_load_no_typecode():
    """test method."""
    # Arrange
    data = json.loads("""
        {
            "title" : "main",
            "typeCode": 2,
            "children": [
                {
                "title" : "title",
                "children": [
                    {
                        "title" : "title1",
                        "uri" : "http://uri1",
                        "annos" : [{
                             "name": "bookmarkProperties/description",
                             "value": "desc"
                         }]
                    }]
                }]
        }""")

    # Act
    items = import_firefox_json(data)

    # Assert
    result = []
    for item in items:
        result.append(item)

    assert len(result) == 0


def test_load_invalid_typecode():
    """test method."""
    # Arrange
    data = json.loads("""
        {
            "title" : "title",
            "children": [
                {
                    "title" : "title1",
                    "typeCode" : 99,
                    "uri" : "http://uri1",
                    "annos" : [{
                         "name": "bookmarkProperties/description",
                         "value": "desc"
                     }]
                }]
        }""")
    # Act
    items = import_firefox_json(data)

    # Assert
    result = []
    for item in items:
        result.append(item)

    assert len(result) == 0

def test_load_folder_with_no_children():
    """test method."""

    # Arrange
    data = json.loads("""
        {
            "title" : "title",
            "typeCode" : 2
        } """)

    # Act
    items = import_firefox_json(data)

    # Assert
    result = []
    for item in items:
        result.append(item)

    assert len(result) == 0

def test_load_one_child():
    """test method."""

    # Arrange
    data = json.loads("""
        {
            "title" : "main",
            "typeCode" : 2,
            "children": [
                {
                "title" : "title",
                "typeCode" : 2,
                "children": [
                    {
                        "title" : "title1",
                        "typeCode" : 1,
                        "uri" : "http://uri1",
                        "annos" : [{
                             "name": "bookmarkProperties/description",
                             "value": "desc"
                          }]
                     }
                ]}
           ]
        }
    """)

    # Act
    items = import_firefox_json(data)

    # Assert
    result = []
    for item in items:
        result.append(item)

    assert len(result) == 1
    assert result[0][0] == 'http://uri1'
    assert result[0][1] == 'title1'
    assert result[0][2] == ','
    assert result[0][3] == 'desc'

def test_load_one_container_child():
    """test method."""

    # Arrange
    data = json.loads("""
        {
            "title" : "main",
            "typeCode": 2,
            "children": [
                {
                "title" : "title",
                "typeCode" : 2,
                "children": [
                    {
                        "title":"bookmark folder",
                        "typeCode":2
                    }]
             }]
         }""")

    # Act
    items = import_firefox_json(data)

    # Assert
    result = []
    for item in items:
        result.append(item)

    assert len(result) == 0

def test_load_many_children():
    """test method."""

    # Arrange
    data = json.loads("""
        {
            "title" : "main",
            "typeCode" : 2,
            "children": [
                {
                "title":"Weitere Lesezeichen",
                "typeCode":2,
                "children": [
                    {"title":"title1","typeCode":1,"uri":"http://uri1.com/#more-74"},
                    {"title":"title2","typeCode":1,"uri":"http://uri2.com/xyz"},
                    {"title":"title3","typeCode":1,"uri":"http://uri3.com"}
                ]}
            ]
        }
    """)

    # Act
    items = import_firefox_json(data)

    # Assert
    result = []
    for item in items:
        result.append(item)

    assert len(result) == 3

def test_load_container_no_title():
    """test method."""

    # Arrange
    data = json.loads("""
        {
            "title" : "main",
            "typeCode" : 2,
            "children": [
                {
                "typeCode" : 2,
                "children": [
                        {"title":"title1","typeCode":1,"uri":"http://uri.com"}
                ]}
            ]
        }
    """)

    # Act
    items = import_firefox_json(data, add_bookmark_folder_as_tag=True)

    # Assert
    result = []
    for item in items:
        result.append(item)

    assert len(result) == 1
    assert result[0][0] == 'http://uri.com'
    assert result[0][2] == ',<no title>,'

def test_load_hierarchical_container_without_ignore():
    """test method."""

    # Arrange
    data = json.loads("""
        {
            "title" : "main",
            "typeCode" : 2,
            "children": [
                {
                "title" : "title",
                "typeCode" : 2,
                "children": [
                     {"title":"title1","typeCode":1,"uri":"http://uri.com"}
                ]
            }]
       }
    """)

    # Act
    items = import_firefox_json(data, add_bookmark_folder_as_tag=True)

    # Assert
    result = []
    for item in items:
        result.append(item)

    assert len(result) == 1
    assert result[0][0] == 'http://uri.com'
    assert result[0][2] == ',title,'

def test_load_hierarchical_container_with_ignore():
    """test method."""

    # Arrange
    data = json.loads("""
        {
            "title" : "main",
            "typeCode" : 2,
            "children": [
                {
                "title" : "title",
                "typeCode" : 2,
                "root": "bookmarksMenuFolder",
                "children": [
                        {
                            "title" : "title2",
                            "typeCode" : 2,
                            "children": [
                                {"title":"title1","typeCode":1,"uri":"http://uri1.com/#more-74"}
                            ]
                        },
                        {"title":"title4","typeCode":1,"uri":"http://uri4.com/#more-74"}
                ]
            }]
       }
    """)

    # Act
    items = import_firefox_json(data, add_bookmark_folder_as_tag=True)

    # Assert
    result = []
    for item in items:
        result.append(item)

    assert len(result) == 2
    assert result[0][0] == 'http://uri1.com/#more-74'
    assert result[1][0] == 'http://uri4.com/#more-74'

    assert result[0][2] == ',title2,'
    assert result[1][2] == ','

def test_load_separator():
    """test method."""

    # Arrange
    data = json.loads("""
        {
            "title" : "main",
            "typeCode" : 2,
            "children": [
                {
                "title" : "title",
                "typeCode" : 2,
                "children": [
                    {
                      "title": "",
                      "type": "text/x-moz-place-separator",
                      "typeCode": 3
                    }]
            }]
         }""")

    # Act
    items = import_firefox_json(data)

    # Assert
    result = []
    for item in items:
        result.append(item)

    assert len(result) == 0

def test_load_multiple_tags():
    """test method."""
    # Arrange
    data = json.loads("""
        {
            "title" : "main",
            "typeCode": 2,
            "children": [
                {
                "title" : "title",
                "typeCode": 2,
                "children": [
                    {
                        "title" : "title1",
                        "uri" : "http://uri1",
                        "tags" : "tag1, tag2",
                        "typeCode": 1,
                        "annos" : [{
                             "name": "bookmarkProperties/description",
                             "value": "desc"
                         }]
                    }]
            }]
        }""")

    # Act
    items = import_firefox_json(data)

    # Assert
    result = []
    for item in items:
        result.append(item)

    assert len(result) == 1
    assert result[0][2] == ",tag1,tag2,"



# clone of flask-reverse-proxy-fix with github fixes from @rachmadaniHaryono and @Glocktober

from flask import Flask as App
# noinspection PyPackageRequirements
try:
    from werkzeug.contrib.fixers import ProxyFix
except ModuleNotFoundError:
    from werkzeug.middleware.proxy_fix import ProxyFix


class ReverseProxyPrefixFix:  # pylint: disable=too-few-public-methods
    """
    Flask middleware to ensure correct URLs are generated by Flask.url_for() where an application is under a reverse
    proxy. Specifically this middleware corrects URLs where a common prefix needs to be added to all URLs.

    For example: If client requests for an application are reverse proxied such that:
    `example.com/some-service/v1/foo` becomes `some-service-v1.internal/foo`, where `/foo` is a route within a Flask
    application `foo()`.

    Without this middleware, a call to `Flask.url_for('.foo')` would give: `/foo`. If returned to the client, as a
    'self' link for example, this would cause a request to `example.com/foo`, which would be invalid as the
    `/some-service/v1` prefix is missing.

    With this middleware, a call to `Flask.url_for('.foo')` would give: '/some-service/v1/foo', which will work if used
    by a client.

    This middleware is compatible with both relative and absolute URLs (i.e. `Flask.url_for('.foo')` and
    `Flask.url_for('.foo', _external=True)`.

    This middleware incorporates the `werkzeug.contrib.fixers.ProxyFix` middleware [1] and is based on the
    'Fixing SCRIPT_NAME/url_scheme when behind reverse proxy' Flask snippet [2].

    Note: Ensure the prefix value includes a preceding slash, but not a trailing slash (i.e. use `/foo` not `/foo/`).

    [1] http://werkzeug.pocoo.org/docs/0.14/contrib/fixers/#werkzeug.contrib.fixers.ProxyFix
    [2] http://flask.pocoo.org/snippets/35/
    """
    def __init__(self, app: App, **kwargs):
        """
        :type app: App
        :param app: Flask application
        """
        self.app = app.wsgi_app
        self.prefix = None

        if 'REVERSE_PROXY_PATH' in app.config:
            self.prefix = app.config['REVERSE_PROXY_PATH']

        self.app = ProxyFix(self.app, **kwargs)

        app.wsgi_app = self

    def __call__(self, environ, start_response):
        if self.prefix is not None:
            environ['SCRIPT_NAME'] = self.prefix
            path_info = environ['PATH_INFO']
            if path_info.startswith(self.prefix):
                environ['PATH_INFO'] = path_info[len(self.prefix):]

        return self.app(environ, start_response)



#!/usr/bin/env python
# pylint: disable=wrong-import-order, ungrouped-imports
"""Server module."""
import collections
import typing as T
from unittest import mock

from flask.views import MethodView

import buku
from buku import BukuDb

import flask
from flask import current_app, redirect, request, url_for

try:
    from response import Response
    from forms import ApiBookmarkCreateForm, ApiBookmarkEditForm, ApiBookmarkRangeEditForm, ApiTagForm
except ImportError:
    from bukuserver.response import Response
    from bukuserver.forms import ApiBookmarkCreateForm, ApiBookmarkEditForm, ApiBookmarkRangeEditForm, ApiTagForm


STATISTIC_DATA = None


def entity(bookmark, id=False):
    data = {
        'id': bookmark.id,
        'url': bookmark.url,
        'title': bookmark.title,
        'tags': bookmark.taglist,
        'description': bookmark.desc,
    }
    if not id:
        data.pop('id')
    return data


def get_bukudb():
    """get bukudb instance"""
    db_file = current_app.config.get('BUKUSERVER_DB_FILE', None)
    return BukuDb(dbfile=db_file)


def search_tag(
    db: BukuDb, stag: T.Optional[str] = None, limit: T.Optional[int] = None
) -> T.Tuple[T.List[str], T.Dict[str, int]]:
    """search tag.

    db:
        buku db instance
    stag:
        search tag
    limit:
        positive integer limit

    Returns
    -------
    tuple
        list of unique tags sorted alphabetically and dictionary of tag and its usage count

    Raises
    ------
    ValueError
        if limit is not positive
    """
    if limit is not None and limit < 1:
        raise ValueError("limit must be positive")
    tags: T.Set[str] = set()
    counter = collections.Counter()
    query_list = ["SELECT DISTINCT tags , COUNT(tags) FROM bookmarks"]
    if stag:
        query_list.append("where tags LIKE :search_tag")
    query_list.append("GROUP BY tags")
    row: T.Tuple[str, int]
    for row in db.cur.execute(" ".join(query_list), {"search_tag": f"%{stag}%"}):
        for tag in row[0].strip(buku.DELIM).split(buku.DELIM):
            if not tag:
                continue
            tags.add(tag)
            counter[tag] += row[1]
    return list(sorted(tags)), dict(counter.most_common(limit))


class ApiTagView(MethodView):

    def get(self, tag: T.Optional[str]):
        bukudb = get_bukudb()
        if tag is None:
            return Response.SUCCESS(data={"tags": search_tag(db=bukudb, limit=5)[0]})
        tags = search_tag(db=bukudb, stag=tag)
        if tag not in tags[1]:
            return Response.TAG_NOT_FOUND()
        return Response.SUCCESS(data={"name": tag, "usage_count": tags[1][tag]})

    def put(self, tag: str):
        form = ApiTagForm({})
        error_response, data = form.process_data(request.get_json())
        if error_response is not None:
            return error_response(data=data)
        bukudb = get_bukudb()
        tags = search_tag(db=bukudb, stag=tag)
        if tag not in tags[1]:
            return Response.TAG_NOT_FOUND()
        try:
            bukudb.replace_tag(tag, form.tags.data)
            return Response.SUCCESS()
        except (ValueError, RuntimeError):
            return Response.FAILURE()

    def delete(self, tag: str):
        if buku.DELIM in tag:
            return Response.TAG_NOT_VALID()
        bukudb = get_bukudb()
        tags = search_tag(db=bukudb, stag=tag)
        if tag not in tags[1]:
            return Response.TAG_NOT_FOUND()
        return Response.from_flag(bukudb.delete_tag_at_index(0, tag, chatty=False))


class ApiBookmarkView(MethodView):

    def get(self, rec_id: T.Union[int, None]):
        if rec_id is None:
            bukudb = getattr(flask.g, 'bukudb', get_bukudb())
            all_bookmarks = bukudb.get_rec_all()
            result = {'bookmarks': [entity(bookmark, id=not request.path.startswith('/api/'))
                                    for bookmark in all_bookmarks]}
        else:
            bukudb = getattr(flask.g, 'bukudb', get_bukudb())
            bookmark = bukudb.get_rec_by_id(rec_id)
            if bookmark is None:
                return Response.BOOKMARK_NOT_FOUND()
            result = entity(bookmark)
        return Response.SUCCESS(data=result)

    def post(self, rec_id: None = None):
        form = ApiBookmarkCreateForm({})
        error_response, error_data = form.process_data(request.get_json())
        if error_response is not None:
            return error_response(data=error_data)
        bukudb = getattr(flask.g, 'bukudb', get_bukudb())
        result_flag = bukudb.add_rec(
            form.url.data,
            form.title.data,
            form.tags_str,
            form.description.data,
            fetch=form.fetch.data)
        return Response.from_flag(result_flag)

    def put(self, rec_id: int):
        form = ApiBookmarkEditForm({})
        error_response, error_data = form.process_data(request.get_json())
        if error_response is not None:
            return error_response(data=error_data)
        bukudb = getattr(flask.g, 'bukudb', get_bukudb())
        result_flag = bukudb.update_rec(
            rec_id,
            form.url.data,
            form.title.data,
            form.tags_str,
            form.description.data)
        return Response.from_flag(result_flag)

    def delete(self, rec_id: T.Union[int, None]):
        if rec_id is None:
            bukudb = getattr(flask.g, 'bukudb', get_bukudb())
            with mock.patch('buku.read_in', return_value='y'):
                result_flag = bukudb.cleardb()
        else:
            bukudb = getattr(flask.g, 'bukudb', get_bukudb())
            result_flag = bukudb.delete_rec(rec_id, retain_order=True)
        return Response.from_flag(result_flag)


class ApiBookmarkRangeView(MethodView):

    def get(self, starting_id: int, ending_id: int):
        bukudb = getattr(flask.g, 'bukudb', get_bukudb())
        max_id = bukudb.get_max_id() or 0
        if starting_id > ending_id or ending_id > max_id:
            return Response.RANGE_NOT_VALID()
        result = {'bookmarks': {i: entity(bukudb.get_rec_by_id(i))
                                for i in range(starting_id, ending_id + 1)}}
        return Response.SUCCESS(data=result)

    def put(self, starting_id: int, ending_id: int):
        bukudb = getattr(flask.g, 'bukudb', get_bukudb())
        max_id = bukudb.get_max_id() or 0
        if starting_id > ending_id or ending_id > max_id:
            return Response.RANGE_NOT_VALID()
        updates = []
        errors = {}
        for rec_id in range(starting_id, ending_id + 1):
            json = request.get_json().get(str(rec_id))
            if json is None:
                errors[rec_id] = 'Input required.'
                continue
            form = ApiBookmarkRangeEditForm({})
            error_response, error_data = form.process_data(json)
            if error_response is not None:
                errors[rec_id] = error_data.get('errors')
            updates += [{'index': rec_id,
                         'url': form.url.data,
                         'title_in': form.title.data,
                         'tags_in': form.tags_in,
                         'desc': form.description.data}]

        if errors:
            return Response.INPUT_NOT_VALID(data={'errors': errors})
        for update in updates:
            if not bukudb.update_rec(**update):
                return Response.FAILURE()
        return Response.SUCCESS()

    def delete(self, starting_id: int, ending_id: int):
        bukudb = getattr(flask.g, 'bukudb', get_bukudb())
        max_id = bukudb.get_max_id() or 0
        if starting_id > ending_id or ending_id > max_id:
            return Response.RANGE_NOT_VALID()
        idx = min([starting_id, ending_id])
        result_flag = bukudb.delete_rec(idx, starting_id, ending_id, is_range=True, retain_order=True)
        return Response.from_flag(result_flag)


class ApiBookmarkSearchView(MethodView):

    def get(self):
        arg_obj = request.args
        keywords = arg_obj.getlist('keywords')
        all_keywords = arg_obj.get('all_keywords')
        deep = arg_obj.get('deep')
        regex = arg_obj.get('regex')
        # api request is more strict
        all_keywords = False if all_keywords is None else all_keywords
        deep = False if deep is None else deep
        regex = False if regex is None else regex
        all_keywords = (
            all_keywords if isinstance(all_keywords, bool) else
            all_keywords.lower() == 'true'
        )
        deep = deep if isinstance(deep, bool) else deep.lower() == 'true'
        regex = regex if isinstance(regex, bool) else regex.lower() == 'true'
        bukudb = getattr(flask.g, 'bukudb', get_bukudb())
        result = {'bookmarks': [entity(bookmark, id=True)
                                for bookmark in bukudb.searchdb(keywords, all_keywords, deep, regex)]}
        current_app.logger.debug('total bookmarks:{}'.format(len(result['bookmarks'])))
        return Response.SUCCESS(data=result)

    def delete(self):
        arg_obj = request.form
        keywords = arg_obj.getlist('keywords')
        all_keywords = arg_obj.get('all_keywords')
        deep = arg_obj.get('deep')
        regex = arg_obj.get('regex')
        # api request is more strict
        all_keywords = False if all_keywords is None else all_keywords
        deep = False if deep is None else deep
        regex = False if regex is None else regex
        all_keywords = (
            all_keywords if isinstance(all_keywords, bool) else
            all_keywords.lower() == 'true'
        )
        deep = deep if isinstance(deep, bool) else deep.lower() == 'true'
        regex = regex if isinstance(regex, bool) else regex.lower() == 'true'
        bukudb = getattr(flask.g, 'bukudb', get_bukudb())
        res = None
        for bookmark in bukudb.searchdb(keywords, all_keywords, deep, regex):
            if not bukudb.delete_rec(bookmark.id, retain_order=True):
                res = Response.FAILURE()
        return res or Response.SUCCESS()


class BookmarkletView(MethodView):  # pylint: disable=too-few-public-methods
    def get(self):
        url = request.args.get('url')
        title = request.args.get('title')
        description = request.args.get('description')
        tags = request.args.get('tags')
        fetch = request.args.get('fetch')

        bukudb = getattr(flask.g, 'bukudb', get_bukudb())
        rec_id = bukudb.get_rec_id(url)
        goto = (url_for('bookmark.edit_view', id=rec_id, popup=True) if rec_id else
                url_for('bookmark.create_view', link=url, title=title, description=description, tags=tags, fetch=fetch, popup=True))
        return redirect(goto)



from unittest import mock
from io import StringIO
import os
import pytest

import buku


@pytest.fixture
def stdin(monkeypatch):
    with monkeypatch.context():
        monkeypatch.setattr('sys.stdin', (buffer := StringIO()))
        yield buffer

@pytest.fixture
def BukuDb():
    with mock.patch('buku.BukuDb') as cls:
        cls.return_value.close_quit.side_effect = SystemExit
        yield cls

@pytest.fixture
def bdb(BukuDb):
    yield BukuDb.return_value

@pytest.fixture
def piped_input():
    with mock.patch('buku.piped_input') as fn:
        yield fn

@pytest.fixture
def prompt():
    with mock.patch('buku.prompt') as fn:
        yield fn

@pytest.fixture
def exit():
    with mock.patch('sys.exit', side_effect=SystemExit) as fn:
        yield fn


def test_version(BukuDb, piped_input, capsys):
    with pytest.raises(SystemExit):
        buku.main(['--version'])
    assert capsys.readouterr().out.splitlines() == [buku.__version__]

def test_usage(BukuDb, piped_input, monkeypatch, capsys):
    with pytest.raises(SystemExit):
        buku.main(['--unknown'], program_name='buku')
    BukuDb.assert_not_called()
    assert capsys.readouterr().err.splitlines() == [
        'usage: buku [OPTIONS] [KEYWORD [KEYWORD ...]]',
        'buku: error: unrecognized arguments: --unknown',
    ]

@pytest.mark.parametrize('argv', [['--help'], ['foo', 'bar', '--help']])
def test_help(BukuDb, exit, piped_input, argv):
    with mock.patch('buku.ExtendedArgumentParser.print_help') as print_help:
        with pytest.raises(SystemExit):
            buku.main(argv)
    BukuDb.assert_not_called()
    print_help.assert_called_with()
    exit.assert_called_with(0)

@pytest.mark.parametrize('nostdin', [True, False])
@pytest.mark.parametrize('db', [None, './foo.db'])
def test_prompt(BukuDb, bdb, piped_input, prompt, nostdin, db):
    argv = (['--nostdin'] if nostdin else []) + (['--db', db] if db else [])
    BukuDb.get_default_dbdir.return_value = '/default/db/dir'
    with pytest.raises(SystemExit):
        buku.main(argv)
    if argv and argv[0] != '--nostdin':
        piped_input.assert_called_with(argv, [])
    else:
        piped_input.assert_not_called()
    BukuDb.assert_called_with(dbfile=db or os.path.join('/default/db/dir', 'bookmarks.db'))
    prompt.assert_called_with(bdb, None)
    bdb.close_quit.assert_called_with(0)


@pytest.mark.parametrize('fetch_params', [
    {'offline': True},
    {'url_redirect': True},
    {'tag_redirect': True, 'tag_error': True},
    {'url_redirect': True, 'tag_redirect': 'redirect', 'tag_error': 'error'},
    {'url_redirect': True, 'tag_redirect': 'redirect', 'del_range': [], 'del_error': range(400, 600)},
    {'url_redirect': True, 'tag_redirect': 'redirect', 'tag_error': 'error',
     'del_range': ['400-404', '500'], 'del_error': {400, 401, 402, 403, 404, 500}},
])
@pytest.mark.parametrize('value_params', [
    {'add_tags': ['foo,bar', 'baz'], 'tags_fetch': False, 'tags_in': ',bar baz,foo,', 'title': ''},
    {'tag': ['foo', 'bar,baz'], 'tags_fetch': False, 'tags_in': ',baz,foo bar,', 'title': 'Custom Title'},
    {'add_tags': ['+', 'foo', 'bar', 'baz'], 'tags_in': ',foo bar baz,', 'comment': ''},
    {'tag': ['+', 'foo,bar,baz'], 'tags_in': ',bar,baz,foo,', 'comment': 'Custom Description'},
    {'add_tags': ['-', 'foo', 'baz', 'baz'], 'tags_except': ',foo baz baz,', 'immutable': False},
    {'tag': ['-', 'foo,', ',baz,', ',baz'], 'tags_except': ',baz,foo,', 'immutable': True},
    {'add_tags': ['foo,baz,bar'], 'tag': ['baz,qux'],
     'tags_fetch': False, 'tags_in': ',bar,baz,foo,qux,'},
    {'add_tags': ['+', 'foo,baz,bar'], 'tag': ['baz,', 'qux,'],
     'tags_fetch': False, 'tags_in': ',bar,baz,foo,qux,'},
    {'add_tags': ['foo,baz,', 'bar,'], 'tag': ['+', 'baz,qux'],
     'tags_fetch': False, 'tags_in': ',bar,baz,foo,qux,'},
    {'add_tags': ['-', 'foo,baz,bar,'], 'tag': ['baz,', 'qux,'],
     'tags_fetch': False, 'tags_in': ',baz,qux,', 'tags_except': ',bar,baz,foo,'},
    {'add_tags': ['foo,baz,', 'bar,'], 'tag': ['-', 'baz,qux'],
     'tags_fetch': False, 'tags_in': ',bar,baz,foo,', 'tags_except': ',baz,qux,'},
    {'add_tags': ['-', 'foo,baz,bar,'], 'tag': ['-', 'baz,', 'qux,'], 'tags_except': ',bar,baz,foo,qux,'},
    {'add_tags': ['-', 'foo,baz,', 'bar,'], 'tag': ['+', 'baz,', 'qux,'],
     'tags_in': ',baz,qux,', 'tags_except': ',bar,baz,foo,'},
    {'add_tags': ['+', 'foo,baz,', 'bar,'], 'tag': ['-', 'baz,', 'qux,'],
     'tags_in': ',bar,baz,foo,', 'tags_except': ',baz,qux,'},
])
def test_add(stdin, bdb, prompt, value_params, fetch_params):
    _test_add(bdb, prompt, **value_params, **fetch_params)

def _test_add(bdb, prompt, *, add_tags=[], tag=[], tags_fetch=True, tags_in=None, tags_except=None,
              title=None, comment=None, immutable=None, offline=False, url_redirect=False,
              tag_redirect=False, tag_error=False, del_range=None, del_error=None):
    argv = ['--add', (url := 'https://example.com/')] + add_tags
    if tag:
        argv += ['--tag'] + tag
    if title is not None:
        argv += ['--title', title]
    if comment is not None:
        argv += ['--comment', comment]
    if immutable is not None:
        argv += ['--immutable', str(int(immutable))]
    if offline:
        argv += ['--offline']
    if url_redirect:
        argv += ['--url-redirect']
    if tag_redirect:
        argv += ['--tag-redirect'] + ([] if isinstance(tag_redirect, bool) else [tag_redirect])
    if tag_error:
        argv += ['--tag-error'] + ([] if isinstance(tag_error, bool) else [tag_error])
    if del_error:
        argv += ['--del-error'] + del_range
    print(argv)
    with pytest.raises(SystemExit):
        buku.main(argv)
    network_test = url_redirect or tag_redirect or tag_error or del_error
    fetch = not offline and (network_test or tags_fetch or title is None)
    bdb.add_rec.assert_called_with(
        url, title, tags_in, comment, immutable, delay_commit=False, fetch=fetch,
        tags_fetch=tags_fetch, tags_except=tags_except, url_redirect=url_redirect,
        tag_redirect=tag_redirect, tag_error=tag_error, del_error=del_error)
    bdb.searchdb.assert_not_called()
    prompt.assert_not_called()
    bdb.close_quit.assert_called_with(0)


@pytest.mark.parametrize('np', [{}, {'np': []}])
@pytest.mark.parametrize('count', [{}, {'count': ['10']}])
@pytest.mark.parametrize('order, indices, command', [
    (['tags', '-netloc', '+url'], None, {'order': ['tags,-netloc,+url'], 'print': []}),
    (['-description', '+uri'], [5, 8, 9, 10, 11, 12, 40, 41, 42],
     {'order': [',-description', '+uri'], 'print': ['5', '8-12', '-3']}),
])
def test_order_print(bdb, stdin, prompt, order, indices, command, count, np):
    command = dict(command, **count, **np)
    argv = [s for k, v in command.items() for s in ([f'--{k}'] + v)]
    print(argv)
    result = [None] * 20
    bdb.list_using_id.return_value = result
    bdb.get_max_id.return_value = 42
    with pytest.raises(SystemExit):
        buku.main(argv)
    bdb.get_max_id.assert_called_with()
    if not (_count := command.get('count')):
        bdb.print_rec.assert_called_with(indices, order=order)
    else:
        if not command['print']:
            bdb.list_using_id.assert_called_with(order=order)
        else:
            bdb.list_using_id.assert_called_with(command['print'], order=order)
        prompt.assert_called_with(bdb, result, noninteractive=('np' in command), num=int(_count[0]), order=order)

@pytest.mark.parametrize('search', ['', 'sany', 'sall', 'sreg', 'stag'])
@pytest.mark.parametrize('exclude', [None, ['xyzzy', 'grue']])
@pytest.mark.parametrize('keywords, rest', [
    ([], {}),
    (['foo', 'bar'], {'markers': []}),
    (['foo', 'bar'], {'deep': []}),
    (['foo', 'bar'], {'stag': ['baz', 'qux']})
])
def test_order_search(bdb, stdin, prompt, search, exclude, keywords, rest):
    if (search == '' and not keywords) or (search == 'stag' and 'stag' in rest):
        pytest.skip('Invalid combination')
    order, stag, deep, markers = ['title', '-index'], rest.get('stag'), 'deep' in rest, 'markers' in rest
    argv = ([] if search == '' else [f'--{search}']) + keywords + ['--order', ','.join(order)]
    argv += [s for k, v in rest.items() for s in ([f'--{k}'] + v)] + ([] if not exclude else ['--exclude'] + exclude)
    bdb.search_by_tag.return_value = ['tag search results']
    with pytest.raises(SystemExit):
        buku.main(argv)
    if search == 'stag':
        if not keywords:
            prompt.assert_called_with(bdb, None, noninteractive=False, listtags=True, suggest=False, order=order)
        else:
            bdb.search_by_tag.assert_called_with(' '.join(keywords), order=order)
            bdb.exclude_results_from_search.assert_called_with(
                bdb.search_by_tag.return_value, exclude, deep=deep, markers=markers)
    if search == 'stag' or not keywords:
        bdb.search_keywords_and_filter_by_tags.assert_not_called()
    elif search in ('', 'sany'):
        bdb.search_keywords_and_filter_by_tags.assert_called_with(
            keywords, deep=deep, stag=stag, markers=markers, without=exclude, order=order)
    elif search == 'sall':
        bdb.search_keywords_and_filter_by_tags.assert_called_with(
            keywords, all_keywords=True, deep=deep, stag=stag, markers=markers, without=exclude, order=order)
    elif search == 'sreg':
        bdb.search_keywords_and_filter_by_tags.assert_called_with(
            keywords, regex=True, stag=stag, markers=markers, without=exclude, order=order)

@pytest.mark.parametrize('json', [None, '', 'output.json'])
@pytest.mark.parametrize('indices', [None, '', '1-10', '-10'])  # None = search
@pytest.mark.parametrize('random', [None, 1, 3])
@mock.patch('random.sample', return_value='sampled')
@mock.patch('buku.print_rec_with_filter')
@mock.patch('buku.write_string_to_file')
@mock.patch('buku.format_json', return_value='formatted')
@mock.patch('buku.print_json_safe')
def test_random(_print_json_safe, _format_json, _write_string_to_file, _print_rec_with_filter, _sample,
                bdb, stdin, prompt, random, indices, json):
    wrap = mock.Mock()
    wrap.attach_mock(_sample, 'random_sample')
    wrap.attach_mock(_print_rec_with_filter, 'print_rec_with_filter')
    wrap.attach_mock(_write_string_to_file, 'write_string_to_file')
    wrap.attach_mock(_format_json, 'format_json')
    wrap.attach_mock(_print_json_safe, 'print_json_safe')
    wrap.attach_mock(prompt, 'prompt')
    wrap.attach_mock(bdb, 'bdb')
    bdb.get_max_id.return_value = 42
    bdb._sort.return_value = 'sorted'
    bdb.search_keywords_and_filter_by_tags.return_value = 'found'
    argv = (['--sall', 'foo'] if indices is None else ['--print'] + ([] if not indices else [indices]))
    argv += ([] if json is None else ['--json'] + ([] if not json else [json]))
    argv += ([] if not random else ['--random'] + ([] if random == 1 else [str(random)]))
    with pytest.raises(SystemExit):
        buku.main(argv)
    calls = ([] if indices is None else [mock.call.bdb.get_max_id()])
    if indices:                # --print 1-10
        idxs = list(range(33, 43) if indices == '-10' else range(1, 11))
        calls += ([mock.call.bdb.print_rec(idxs, order=[])] if not random else
                  [mock.call.random_sample(idxs, random),
                   mock.call.bdb.print_rec('sampled', order=[])])
    elif indices is not None:  # --print
        calls += ([mock.call.bdb.print_rec(None, order=[])] if not random else
                  [mock.call.random_sample(range(1, 43), random),
                   mock.call.bdb.print_rec('sampled', order=[])])
    else:                      # --sall foo
        calls += [mock.call.bdb.search_keywords_and_filter_by_tags(
                      ['foo'], all_keywords=True, deep=False, stag=None, markers=False, without=None, order=[])]
        if random:
            calls += [mock.call.random_sample('found', random),
                      mock.call.bdb._sort('sampled', [])]
        res = ('sorted' if random else 'found')
        if json:
            calls += [mock.call.format_json(res, (random == 1), field_filter=0),
                      mock.call.write_string_to_file('formatted', json)]
        elif json is not None:
            calls += [mock.call.print_json_safe(res, (random == 1), field_filter=0)]
        elif random:
            calls += [mock.call.print_rec_with_filter(res, field_filter=0)]
        else:
            calls += [mock.call.prompt(bdb, res, noninteractive=False, deep=False, markers=False, order=[], num=10)]
    calls += [mock.call.bdb.close_quit(0)]
    assert wrap.mock_calls == calls

@pytest.mark.parametrize('search', [True, False])
@pytest.mark.parametrize('random', [None, 1, 3])
@mock.patch('random.sample', return_value='sampled')
@mock.patch('buku.print_rec_with_filter')
def test_random_export(_print_rec_with_filter, _sample, bdb, stdin, prompt, random, search):
    wrap = mock.Mock()
    wrap.attach_mock(_sample, 'random_sample')
    wrap.attach_mock(_print_rec_with_filter, 'print_rec_with_filter')
    wrap.attach_mock(prompt, 'prompt')
    wrap.attach_mock(bdb, 'bdb')
    bdb.get_max_id.return_value = 42
    bdb._sort.return_value = 'sorted'
    bdb.search_keywords_and_filter_by_tags.return_value = 'found'
    argv = ['--export', 'export.md'] + ([] if not search else ['--sall', 'foo'])
    argv += ([] if not random else ['--random'] + ([] if random == 1 else [str(random)]))
    with pytest.raises(SystemExit):
        buku.main(argv)
    calls = []
    if not search:
        calls += [mock.call.bdb.exportdb('export.md', order=[], pick=random)]
    else:
        calls += [mock.call.bdb.search_keywords_and_filter_by_tags(
                      ['foo'], all_keywords=True, deep=False, stag=None, markers=False, without=None, order=[])]
        if random:
            calls += [mock.call.random_sample('found', random),
                      mock.call.bdb._sort('sampled', [])]
        res = ('sorted' if random else 'found')
        if random:
            calls += [mock.call.print_rec_with_filter(res, field_filter=0)]
        else:
            calls += [mock.call.prompt(bdb, res, noninteractive=True, deep=False, markers=False, order=[], num=10)]
        calls += [mock.call.bdb.exportdb('export.md', res)]
    calls += [mock.call.bdb.close_quit(0)]
    assert wrap.mock_calls == calls


@pytest.mark.parametrize('db', [None, './foo.db', 'bar.sqlite', 'name'])
@pytest.mark.parametrize('action', ['print', 'lock', 'unlock'])
@mock.patch('buku.BukuCrypt')
def test_custom_db(_BukuCrypt, BukuDb, stdin, db, action):
    wrap = mock.Mock()
    wrap.attach_mock(BukuDb, 'BukuDb')
    wrap.attach_mock(BukuDb.return_value, 'bdb')
    wrap.attach_mock(_BukuCrypt, 'BukuCrypt')
    BukuDb.return_value.get_max_id.return_value = None
    BukuDb.get_default_dbdir.return_value = '/default/db/dir'
    _db = (db if db != 'name' else '/default/db/dir/name.db')
    argv = ['--nostdin'] + ([] if not db else ['--db', db]) + [f'--{action}']
    with pytest.raises(SystemExit):
        buku.main(argv)
    calls = []
    if db == 'name':
        calls += [mock.call.BukuDb.get_default_dbdir()]
    if action == 'lock':
        calls += [mock.call.BukuCrypt.encrypt_file(8, dbfile=_db)]
    elif action == 'unlock':
        calls += [mock.call.BukuCrypt.decrypt_file(8, dbfile=_db)]
    calls += [mock.call.BukuDb(None, 0, True, dbfile=_db, colorize=True)]
    if action == 'print':
        calls += [mock.call.bdb.get_max_id(),
                  mock.call.bdb.print_rec(None, order=[])]
    calls += [mock.call.bdb.close_quit(0)]
    assert wrap.mock_calls == calls



"""Forms module."""
# pylint: disable=too-few-public-methods, missing-docstring
from typing import Any, Dict, Tuple
from flask_wtf import FlaskForm
from wtforms.fields import BooleanField, FieldList, StringField, TextAreaField, HiddenField
from wtforms.validators import DataRequired, InputRequired, ValidationError
from buku import DELIM, parse_tags
from bukuserver import _, _l, LazyString
from bukuserver.response import Response

def validate_tag(form, field):
    if not isinstance(field.data, str):
        raise ValidationError(_('Tag must be a string.'))
    if DELIM in field.data:
        raise ValidationError(_('Tag must not contain delimiter "%(delim)s".', delim=DELIM))


class SearchBookmarksForm(FlaskForm):
    keywords = FieldList(StringField(_l('Keywords')), min_entries=1)
    all_keywords = BooleanField(_l('Match all keywords'), default=True, description=_l('Exclude partial matches (with multiple keywords)'))
    markers = BooleanField(_l('With markers'), default=True, description=LazyString(lambda: '\n'.join([
        _('The search string will be split into multiple keywords, each will be applied to a field based on prefix:'),
        _(" - keywords starting with '.', '>' or ':' will be searched for in title, description and URL respectively"),
        _(" - '#' will be searched for in tags (comma-separated, partial matches; not affected by Deep Search)"),
        _(" - '#,' is the same but will match FULL tags only"),
        _(" - '*' will be searched for in all fields (this prefix can be omitted in the 1st keyword)"),
        _('Keywords need to be separated by placing spaces before the prefix.'),
    ])))
    deep = BooleanField(_l('Deep search'), description=_l('When unset, only FULL words will be matched.'))
    regex = BooleanField(_l('Regex'), description=_l('The keyword(s) are regular expressions (overrides other options).'))


class HomeForm(SearchBookmarksForm):
    keyword = StringField(_l('Keyword'))


class BookmarkForm(FlaskForm):
    url = StringField(_l('URL'), name='link', validators=[InputRequired()])
    title = StringField(_l('Title'))
    tags = StringField(_l('Tags'))
    description = TextAreaField(_l('Description'))
    fetch = HiddenField(filters=[bool])


class SwapForm(FlaskForm):
    id1 = HiddenField(filters=[int])
    id2 = HiddenField(filters=[int])


class ApiTagForm(FlaskForm):
    class Meta:
        csrf = False

    tags = FieldList(StringField(validators=[DataRequired(), validate_tag]), min_entries=1)

    tags_str = None

    def process_data(self, data: Dict[str, Any]) -> Tuple[Response, Dict[str, Any]]:
        """Generate comma-separated string tags_str based on list of tags."""
        tags = data.get('tags')
        if tags and not isinstance(tags, list):
            return Response.INPUT_NOT_VALID, {'errors': {'tags': _('List of tags expected.')}}

        super().process(data=data)
        if not self.validate():
            return Response.INPUT_NOT_VALID, {'errors': self.errors}

        self.tags_str = None if tags is None else parse_tags([DELIM.join(tags)])
        return None, None


class ApiBookmarkCreateForm(ApiTagForm):
    class Meta:
        csrf = False

    url = StringField(validators=[DataRequired()])
    title = StringField()
    description = StringField()
    tags = FieldList(StringField(validators=[validate_tag]), min_entries=0)
    fetch = HiddenField(filters=[bool], default=True)


class ApiBookmarkEditForm(ApiBookmarkCreateForm):
    url = StringField()


class ApiBookmarkRangeEditForm(ApiBookmarkEditForm):

    del_tags = BooleanField(_('Delete tags list from existing tags'), default=False)

    tags_in = None

    def process_data(self, data: Dict[str, Any]) -> Tuple[Response, Dict[str, Any]]:
        """Generate comma-separated string tags_in based on list of tags."""
        error_response, data = super().process_data(data)

        if self.tags_str is not None:
            self.tags_in = ("-" if self.del_tags.data else "+") + self.tags_str

        return error_response, data



"""test module."""
from itertools import product
from unittest import mock

import pytest


@pytest.mark.parametrize("platform, file", product(['win32', 'linux'], [None, mock.Mock()]))
def test_program_info(platform, file):
    """test method."""
    with mock.patch('buku.sys') as m_sys:
        import buku
        file = mock.Mock()
        if file is None:
            buku.ExtendedArgumentParser.program_info()
        else:
            buku.ExtendedArgumentParser.program_info(file)
        if platform == 'win32' and file == m_sys.stdout:
            assert len(m_sys.stderr.write.mock_calls) == 1
        else:
            assert len(file.write.mock_calls) == 1


def test_prompt_help():
    """test method."""
    file = mock.Mock()
    import buku
    buku.ExtendedArgumentParser.prompt_help(file)
    assert len(file.write.mock_calls) == 1


def test_print_help():
    """test method."""
    file = mock.Mock()
    import buku
    obj = buku.ExtendedArgumentParser()
    obj.program_info = mock.Mock()
    obj.print_help(file)
    obj.program_info.assert_called_once_with(file)



from unittest import mock
import os

from urllib3 import HTTPResponse

from buku import FetchResult


def mock_http(body=None, **kwargs):
    body = (None if not body else str(body).encode('UTF-8'))
    return mock.patch('urllib3.PoolManager.request', return_value=HTTPResponse(body, **kwargs))

def mock_fetch(custom=None, **kwargs):
    _url = kwargs.pop('url', None)
    status = kwargs.pop('fetch_status', (None if kwargs.get('bad') else 200))
    fn = lambda url, http_head=False: FetchResult(_url or url, fetch_status=status, **kwargs)
    return mock.patch('buku.fetch_data', side_effect=custom or fn)

def _add_rec(db, *args, **kw):
    """Use THIS instead of db.add_rec() UNLESS you want to wait for unnecessary network requests."""
    return db.add_rec(*args, fetch=False, **kw)

def _tagset(s):
    return set(x for x in str(s or '').lower().split(',') if x)

def append(buffer, text):
    pos = buffer.tell()
    try:
        buffer.seek(0, os.SEEK_END)
        return buffer.write(text)
    finally:
        buffer.seek(pos)



#!/usr/bin/env python3

import os
import re
import shutil

from setuptools import find_packages, setup

if os.path.isfile('buku'):
    shutil.copyfile('buku', 'buku.py')

with open('buku.py', encoding='utf-8') as f:
    version = re.search('__version__ = \'([^\']+)\'', f.read()).group(1)  # type: ignore

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

tests_require = [
    'attrs>=17.4.0',
    'beautifulsoup4>=4.6.0',
    'Click>=7.0',
    'flake8>=3.4.1',
    'hypothesis>=6.0.0',
    'mypy-extensions==0.4.1',
    'py>=1.5.0',
    'pylint>=1.7.2',
    'pytest-cov',
    'pytest-recording>=0.12.1',
    'pytest-timeout',
    'pytest>=6.2.1',
    'PyYAML>=4.2b1',
    'setuptools>=41.0.1',
    'vcrpy>=1.13.0',
    'lxml',
    'flask_babel',
]


server_require = [
    "arrow>=1.2.2",
    "Flask-Admin>=1.6.1,<2",
    "flask-paginate>=2022.1.8",
    "Flask-WTF>=1.0.1",
    "Flask>=2.2.2",
    "Jinja2>=3",
]
install_requires = [
    'beautifulsoup4>=4.4.1',
    'certifi',
    'cryptography>=1.2.3',
    'html5lib>=1.0.1',
    'urllib3>=1.23,<3',
    'pyreadline3; sys_platform == \'win32\'',
    'colorama>=0.4.6; sys_platform == \'win32\'',
]

setup(
    name='buku',
    version=version,
    description='Bookmark manager like a text-based mini-web.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Arun Prakash Jana',
    author_email='engineerarun@gmail.com',
    url='https://github.com/jarun/buku',
    license='GPLv3',
    python_requires='>=3.8',  # requires pip>=9.0.0
    platforms=['any'],
    py_modules=['buku'],
    install_requires=install_requires,
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    entry_points={
        'console_scripts': ['buku=buku:main', 'bukuserver=bukuserver.server:cli']
    },
    extras_require={
        "tests": tests_require + server_require,
        "server": server_require,
        "docs": [
            "myst-parser>=0.17.0",
            "sphinx-rtd-theme>=1.0.0",
            "sphinx-autobuild>=2021.3.14",
        ],
        "packaging": ["twine>=1.11.0"],
    },
    test_suite='tests',
    tests_require=tests_require,
    keywords='cli bookmarks tag utility',
    project_urls={
        "Documentation": "https://buku.readthedocs.io/en/latest",
        "Funding": "https://github.com/sponsors/jarun",
        "Source": "https://github.com/jarun/buku",
        "Tracker": "https://github.com/jarun/buku/issues",
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Utilities'
    ]
)
